package io.tswf.selenium.undetected.chrome.driver

import groovy.transform.CompileStatic
import io.tswf.selenium.undetected.chrome.driver.factory.UndetectedChromeDriverBuilder
import io.tswf.selenium.undetected.chrome.driver.factory.UndetectedChromeDriverBuilderImpl
import io.tswf.selenium.undetected.chrome.process.ChromeBrowserProcessStarterImpl
import org.openqa.selenium.Capabilities
import org.openqa.selenium.chrome.ChromeDriver
import org.openqa.selenium.chrome.ChromeOptions

@CompileStatic
class UndetectedChromeDriver extends ChromeDriver {

    boolean headless
    boolean keepUserDataDir
    File userDataDir
    ChromeOptions chromeOptions
    Process remoteBrowserProcess

    UndetectedChromeDriver(
            ChromeOptions chromeOptions,
            boolean headless,
            File userDataDir,
            Process remoteBrowserProcess
    ) {

        super(chromeOptions)
        this.chromeOptions = chromeOptions
        this.headless = headless
        this.userDataDir = userDataDir
        this.remoteBrowserProcess = remoteBrowserProcess
    }

    @Override
    void get(String url) {
        if (headless) {
            _headless()
        }
        removeCdc()
        super.get(url)
    }

    @Override
    void quit() {
        super.quit()

        remoteBrowserProcess.destroyForcibly()

        //delete temp user dir
        if (keepUserDataDir) {
            for (int i = 0; i < 5; i++) {
                try {
                    if (!userDataDir.exists()) {
                        break
                    }
                    userDataDir.deleteIfExists()
                    break
                }
                catch (Exception e) {
                    try {
                        Thread.sleep(300)
                    } catch (Exception ignored) {
                    }
                }
            }
        }

    }

    /**
     * configure headless
     */
    private void _headless() {
        //set navigator.webdriver
        Object f = this.executeScript("return navigator.webdriver")
        if (f == null) {
            return
        }

        Map<String, Object> params1 = new HashMap<>()
        params1.put("source", "Object.defineProperty(window, 'navigator', {\n" +
                "    value: new Proxy(navigator, {\n" +
                "        has: (target, key) => (key === 'webdriver' ? false : key in target),\n" +
                "        get: (target, key) =>\n" +
                "            key === 'webdriver' ?\n" +
                "            false :\n" +
                "            typeof target[key] === 'function' ?\n" +
                "            target[key].bind(target) :\n" +
                "            target[key]\n" +
                "        })\n" +
                "})")

        this.executeCdpCommand("Page.addScriptToEvaluateOnNewDocument", params1)

        //set ua
        Map<String, Object> params2 = new HashMap<>()
        params2.put("userAgent", ((String) this.executeScript("return navigator.userAgent")).replace("Headless", ""))
        this.executeCdpCommand("Network.setUserAgentOverride", params2)

        Map<String, Object> params3 = new HashMap<>()
        params3.put("source", "Object.defineProperty(navigator, 'maxTouchPoints', {get: () => 1})")
        this.executeCdpCommand("Page.addScriptToEvaluateOnNewDocument", params3)

        Map<String, Object> params4 = new HashMap<>()
        params4.put("source", "" +
                "Object.defineProperty(navigator.connection, 'rtt', {get: () => 100})\n" +
                "// https://github.com/microlinkhq/browserless/blob/master/packages/goto/src/evasions/chrome-runtime.js\n" +
                "window.chrome = {\n" +
                "        app: {\n" +
                "            isInstalled: false,\n" +
                "            InstallState: {\n" +
                "                DISABLED: 'disabled',\n" +
                "                INSTALLED: 'installed',\n" +
                "                NOT_INSTALLED: 'not_installed'\n" +
                "            },\n" +
                "            RunningState: {\n" +
                "                CANNOT_RUN: 'cannot_run',\n" +
                "                READY_TO_RUN: 'ready_to_run',\n" +
                "                RUNNING: 'running'\n" +
                "            }\n" +
                "        },\n" +
                "        runtime: {\n" +
                "            OnInstalledReason: {\n" +
                "                CHROME_UPDATE: 'chrome_update',\n" +
                "                INSTALL: 'install',\n" +
                "                SHARED_MODULE_UPDATE: 'shared_module_update',\n" +
                "                UPDATE: 'update'\n" +
                "            },\n" +
                "            OnRestartRequiredReason: {\n" +
                "                APP_UPDATE: 'app_update',\n" +
                "                OS_UPDATE: 'os_update',\n" +
                "                PERIODIC: 'periodic'\n" +
                "            },\n" +
                "            PlatformArch: {\n" +
                "                ARM: 'arm',\n" +
                "                ARM64: 'arm64',\n" +
                "                MIPS: 'mips',\n" +
                "                MIPS64: 'mips64',\n" +
                "                X86_32: 'x86-32',\n" +
                "                X86_64: 'x86-64'\n" +
                "            },\n" +
                "            PlatformNaclArch: {\n" +
                "                ARM: 'arm',\n" +
                "                MIPS: 'mips',\n" +
                "                MIPS64: 'mips64',\n" +
                "                X86_32: 'x86-32',\n" +
                "                X86_64: 'x86-64'\n" +
                "            },\n" +
                "            PlatformOs: {\n" +
                "                ANDROID: 'android',\n" +
                "                CROS: 'cros',\n" +
                "                LINUX: 'linux',\n" +
                "                MAC: 'mac',\n" +
                "                OPENBSD: 'openbsd',\n" +
                "                WIN: 'win'\n" +
                "            },\n" +
                "            RequestUpdateCheckStatus: {\n" +
                "                NO_UPDATE: 'no_update',\n" +
                "                THROTTLED: 'throttled',\n" +
                "                UPDATE_AVAILABLE: 'update_available'\n" +
                "            }\n" +
                "        }\n" +
                "}\n" +
                "\n" +
                "// https://github.com/microlinkhq/browserless/blob/master/packages/goto/src/evasions/navigator-permissions.js\n" +
                "if (!window.Notification) {\n" +
                "        window.Notification = {\n" +
                "            permission: 'denied'\n" +
                "        }\n" +
                "}\n" +
                "\n" +
                "const originalQuery = window.navigator.permissions.query\n" +
                "window.navigator.permissions.__proto__.query = parameters =>\n" +
                "        parameters.name === 'notifications'\n" +
                "            ? Promise.resolve({ state: window.Notification.permission })\n" +
                "            : originalQuery(parameters)\n" +
                "        \n" +
                "const oldCall = Function.prototype.call \n" +
                "function call() {\n" +
                "        return oldCall.apply(this, arguments)\n" +
                "}\n" +
                "Function.prototype.call = call\n" +
                "\n" +
                "const nativeToStringFunctionString = Error.toString().replace(/Error/g, 'toString')\n" +
                "const oldToString = Function.prototype.toString\n" +
                "\n" +
                "function functionToString() {\n" +
                "        if (this === window.navigator.permissions.query) {\n" +
                "            return 'function query() { [native code] }'\n" +
                "        }\n" +
                "        if (this === functionToString) {\n" +
                "            return nativeToStringFunctionString\n" +
                "        }\n" +
                "        return oldCall.call(oldToString, this)\n" +
                "}\n" +
                "// eslint-disable-next-line\n" +
                "Function.prototype.toString = functionToString")
        this.executeCdpCommand("Page.addScriptToEvaluateOnNewDocument", params4)


    }

    /**
     * remove cdc
     */
    private void removeCdc() {
        List<String> f = (List<String>) this.executeScript("let objectToInspect = window,\n" +
                "    result = []\n" +
                "while(objectToInspect !== null)\n" +
                "{ result = result.concat(Object.getOwnPropertyNames(objectToInspect))\n" +
                "  objectToInspect = Object.getPrototypeOf(objectToInspect) }\n" +
                "return result.filter(i => i.match(/.+_.+_(Array|Promise|Symbol)/ig))")

        if (f != null && f.size() > 0) {
            Map<String, Object> param = new HashMap<>()
            param.put("source", "let objectToInspect = window,\n" +
                    "    result = []\n" +
                    "while(objectToInspect !== null)\n" +
                    "{ result = result.concat(Object.getOwnPropertyNames(objectToInspect))\n" +
                    "  objectToInspect = Object.getPrototypeOf(objectToInspect) }\n" +
                    "result.forEach(p => p.match(/.+_.+_(Array|Promise|Symbol)/ig)\n" +
                    "                    &&delete window[p]&&console.log('removed',p))")
            this.executeCdpCommand("Page.addScriptToEvaluateOnNewDocument", param)
        }

    }

    /**
     *  set stealth
     */
    private void _stealth() {
        this.class.classLoader.getResourceAsStream("js/stealth.min.js")?.getText("UTF-8")?.let { stealthJsContent ->
            this.executeCdpCommand("Page.addScriptToEvaluateOnNewDocument", [ source: (Object) stealthJsContent])
        }
    }

    @Override
    void startSession(Capabilities capabilities) {
        if (capabilities == null) {
            capabilities = this.chromeOptions
        }
        super.startSession(capabilities)
    }

    static UndetectedChromeDriverBuilder builder(File driverBinaryLocation, File chromeBrowserBinaryLocation) {
        return new UndetectedChromeDriverBuilderImpl(driverBinaryLocation, chromeBrowserBinaryLocation, new ChromeBrowserProcessStarterImpl())
    }
}
