package io.tswf.selenium.undetected.chrome.patcher

import groovy.transform.CompileStatic

import java.nio.charset.StandardCharsets
import java.util.regex.Pattern

interface ChromeDriverUndetectedPatcher {
    boolean isDriverExecutablePatched(File driverExecutableFile)
    void patchDriverExecutable(File driverExecutableFile)
}

@CompileStatic
class ChromeDriverUndetectedPatcherImpl implements ChromeDriverUndetectedPatcher {

    @Override
    boolean isDriverExecutablePatched(File driverExecutableFile) {
        return driverExecutableFile.newReader("ISO_8859_1").doThenClose { br ->
            def line = (String) null
            
            while ((line = br.readLine()) != null) {
                if (line.contains("undetected chromedriver")) {
                    return true
                }
            }

            return false
        }
    }

    @Override
    void patchDriverExecutable(File driverExecutableFile) {
        new RandomAccessFile(driverExecutableFile, "rw").doThenClose { file ->
            def fileContent = driverExecutableFile.getText("ISO_8859_1")

            def cdcReplacementMatcher = Pattern.compile("\\{window\\.cdc.*?\\}").matcher(fileContent)

            if(cdcReplacementMatcher.find()){
                def group = cdcReplacementMatcher.group()
                def newTarget = new StringBuilder("{console.log(\"undetected chromedriver!\"}")
                int k = group.length() - newTarget.length()
                for (int i = 0; i < k; i++) {
                    newTarget.append(" ")
                }

                def newContent = fileContent.replace(group, newTarget.toString())

                file.seek(0)
                file.write(newContent.getBytes(StandardCharsets.ISO_8859_1))
            }
        }
    }
}
